(in-package :cl-matrix)

;; PASSWORD HASHING

(defun generate-hash (passphrase salt)
  "Where PASSPHRASE is a string and SALT is a unsigned byte array or string, returns an unsigned byte array"
  (let* ((salt (if (typep salt '(simple-array (unsigned-byte 8) (*)))
                   salt
                   (ironclad:hex-string-to-byte-array salt)))
         (kdf (ironclad:make-kdf :argon2id :block-count 19456))
         (passphrase-bytes (flexi-streams:string-to-octets passphrase :external-format :utf-8))
         (hash (ironclad:derive-key kdf passphrase-bytes salt 2 16)))
    hash))

(defun generate-hash-with-salt (passphrase)
  "Where PASSPHRASE is a string, returns a string '<hash>:<salt>'"
  (let* ((salt (ironclad:make-random-salt))
         (hash (generate-hash passphrase salt)))
    (format nil "~a:~a"
            (ironclad:byte-array-to-hex-string hash)
            (ironclad:byte-array-to-hex-string salt))))

(defun check-password (to-check passphrase-hash-and-salt-string)
  (let* ((split-passphrase-hash-and-salt (uiop:split-string
                                          passphrase-hash-and-salt-string :separator ":"))
         (passphrase-hash (first split-passphrase-hash-and-salt))
         (passphrase-salt (second split-passphrase-hash-and-salt))
         (to-check-hash (ironclad:byte-array-to-hex-string (generate-hash to-check passphrase-salt))))
    (equal to-check-hash passphrase-hash)))

;; AUTH SESSION

(defvar *auth-sessions* (make-hash-table :test 'equal))

(defclass auth-session ()
  ((stages :initarg :stages :accessor stages)
   (completed :initform '() :accessor completed)
   (params :initarg :params :initform '() :accessor params)
   (session-id :initform (frugal-uuid:to-string (frugal-uuid:make-v4)) :reader session-id)))

(defmethod auth-session-plist ((auth-session auth-session))
  (let ((plist (list :flows `(:stages ,(stages auth-session))
                     :params (params auth-session)
                     :session (session-id auth-session))))
    (when (completed auth-session)
      (push (completed auth-session) plist)
      (push :completed plist))
    plist))

(defclass password-auth-session (auth-session)
  ((stages :initform (list (make-instance 'password-auth-stage)))))

;;

;; AUTH STAGE
(defclass auth-stage ()
  ((type :reader auth-stage-type)))

(defclass password-auth-stage (auth-stage)
  ((type :initform "m.login.password.")))

(defgeneric process-input (auth-stage input)
  (:documentation
   "Process the plist INPUT as specified by AUTH-STAGE. \

Return two values, the first being T if the input is successful, NIL if not. \
The second value is when the first is NIL and there's additional information regarding the failure. \
We're not signaling any conditions from here because depending on whether it's a user-interactive authorization or plain login will affect response."))

(defmethod process-input ((auth-stage (eql :|m.login.password|)) input)
  (destructuring-bind (&key identifier password &allow-other-keys) input
    (let* ((user (get-user identifier))
           (password-hash (password-hash user)))
      (check-password password password-hash))))

(defmethod process-input (auth-stage input)
  (error 'unknown-error :msg (format nil "invalid auth type: ~a" auth-stage)))

;;

(defun get-user-from-access-token (access-token)
  (let ((device (get-device access-token)))
    (if device
        (get-user (user-id device)))))

(defmacro with-authentication (input-var &body body)
  `(let* ((access-token (matrix-input-access-token ,input-var))
          (user (if access-token (get-user-from-access-token access-token))))
     (setf (matrix-input-user ,input-var))
     (if user
         (progn
           ,@body)
         (handle-auth-session))))

(defun get-auth-params (matrix-input)
  (getf (matrix-input-json-body matrix-input) :auth))

(defun handle-auth-session (matrix-input)
  (if (get-auth-params matrix-input)
      (continue-auth-session matrix-input)
      (start-auth-session matrix-input)))

;; (defun start-auth-session (matrix-input)
;;   (let ((auth-session (make-instance 'password-auth-session)))
;;     (setf (gethash (session-id auth-session) *auth-sessions*) auth-session)
;;     (json-response 401 (auth-session-plist auth-session))))

;; ENDPOINTS

(defun get-login (input)
  (declare (ignore input))
  (json-response 200 '(:flows ((:type "m.login.password")))))

(defun post-login (input)
  (let ((json-body (matrix-input-json-body input)))
    (if (process-input (alex:make-keyword (getf json-body :type "")) json-body)
        (json-response 200 '(:msg "success!"))
        (error 'forbidden-error))))
                            

    
