(in-package :cl-matrix)

(defun get-access-token-from-params (params-alist)
  (cdr (assoc "access_token" params-alist :test 'equal)))

(defun get-access-token-from-header (header)
  (when header
    (aref (nth-value 1 (ppcre:scan-to-strings "^Bearer (.+)$" header)) 0)))
  
(defun get-access-token (env)
  (destructuring-bind (&key headers query-string &allow-other-keys) env
    (let* ((auth-token-header (get-access-token-from-header (gethash "Authorization" headers)))
           (auth-token-params (get-access-token-from-params (if query-string (quri:url-decode-params query-string)))))
      (or auth-token-header auth-token-params))))

(defun received-method-p (method)
  (or (eql method :post)
      (eql method :put)))

(defun initialize-matrix-input (path-args env)
  (destructuring-bind (&key raw-body request-method &allow-other-keys) env
    (make-matrix-input
     :path-args path-args
     :access-token (get-access-token env)
     :json-body (if (received-method-p request-method) (json-plist raw-body) nil))))

(defvar +matrix-headers+
  '(:access-control-allow-origin "*"
    :access-control-allow-methods "GET, POST, PUT, DELETE, OPTIONS"
    :access-control-allow-headers "X-Requested-With, Content-Type, Authorization"))

(defun json-response (code plist &optional headers)
  `(,code ,(append '(:content-type "application/json") headers +matrix-headers+) (,(plist-json plist))))

(defgeneric handle-condition (condition))

(defmethod handle-condition (condition)
  (json-response 500 `(:error ,(format nil "~a" condition))))

(defmethod handle-condition ((err matrix-error))
  (json-response 400 `(:errcode ,(errcode err) :error ,(msg err))))

(defmethod handle-condition ((err unrecognized-error))
  (json-response 404 `(:errcode ,(errcode err) :error ,(msg err))))

(defun wrap-url-match (url-pattern handler)
  (lambda (env)
    (let ((matches (nth-value 1 (ppcre:scan-to-strings url-pattern (getf env :path-info)))))
      (if matches
          (apply handler env (coerce matches 'list))))))

(defun wrap-method-match (method handler)
  (lambda (env)
    (if (eql method (getf env :request-method))
        (funcall handler env))))

(defparameter *routes* nil)

(defmacro defroute (method url-pattern handler)
  `(progn
     (push 
      (wrap-method-match
       ,method
       (wrap-url-match
        ,url-pattern
        (lambda (env &rest rest)
          ;; innermost lambda expects env and path-args for rest
          (funcall (function ,handler) (initialize-matrix-input rest env)))))
      *routes*)
     (push
      (wrap-method-match
       :options
       (wrap-url-match
        ,url-pattern
        (lambda (env &rest rest)
          (json-response 200 '()))))
      *routes*)))

(defun router (env)
  (log-info (getf env :path-info))
  (handler-case
      (loop for route in *routes*
            when (funcall route env)
              return it
            finally (error 'unrecognized-error))
    (matrix-error (c)
      (handle-condition c))))
      

;; earlier defined routes have lower precedence
(defroute :get "^/_matrix/client/versions$" versions)
(defroute :get "^/_matrix/client/v3/login" get-login)
(defroute :post "^/_matrix/client/v3/login" post-login)
  
